import Utils._
import Tseytin._
import Expr._
import scala.util.Random
import scala.io.Source
import java.io.File

case class VarId(id: Int) extends AnyVal {
  def toLiteral(value: Boolean): Lit = Lit(id * bool2Int(value))

  override def toString: String = id.toString
}

case class Lit(lit: Int) extends AnyVal {
  def unary_!(): Lit = Lit(-lit)
  def unary_-(): Lit = Lit(-lit)
  def getVar: VarId = VarId(math.abs(lit))
  def isPositive: Boolean = lit > 0
  def positiveValue: Int = bool2Int(isPositive)

  def toExpr: Expr =
    if (isPositive) Var(StateIdentifier(s"x$lit"))
    else Not(Var(StateIdentifier(s"x${-lit}")))

  override def toString: String = lit.toString
}

case class Clause(literals: Array[Lit]) {
  def maxId: Int = literals.map(_.getVar.id).max
  def isEmpty: Boolean = literals.isEmpty
  def ++(other: Clause): Clause = Clause(literals ++ other.literals)

  def toExpr: Expr = or(literals.map(_.toExpr))

  override def toString: String = {
    literals.mkString(" ")
  }
}

case class CNF(clauses: Array[Clause], name: String = "") {
  def ++(other: CNF): CNF = CNF(clauses ++ other.clauses)
  def +(clause: Clause): CNF = CNF(clauses :+ clause)
  def maxId: Int = clauses.map(_.maxId).max

  def toExpr: Expr = and(clauses.map(_.toExpr))

  override def toString: String = {
    clauses.mkString("\n")
  }
}

object CNF {

  def apply(clauses: Clause*): CNF = CNF(clauses.toArray)

  def makeCNF(e: Expr): (CNF, Map[Identifier, VarId]) = {
    val idToVarId: Map[Identifier, VarId] =
      e.fv.zipWithIndex.toMap.mapValues((x: Int) => VarId(x + 1))

    def topLevelLoop(e: Expr): CNF = e match {
      case And(e1, e2) => topLevelLoop(e1) ++ topLevelLoop(e2)
      case Or(_, _) => CNF(innerLoop(e))
      case Var(_) => CNF(innerLoop(e))
      case Not(Var(_)) => CNF(innerLoop(e))
      case _ => throw new Exception(s"(Building CNF) Expression $e is not in CNF.")
    }

    def innerLoop(e: Expr): Clause = e match {
      case Or(e1, e2) => innerLoop(e1) ++ innerLoop(e2)
      case Var(id) => Clause(Array(idToVarId(id).toLiteral(true)))
      case Not(Var(id)) => Clause(Array(idToVarId(id).toLiteral(false)))
      case _ => throw new Exception(s"(Building Clause) Expression $e is not in CNF.")
    }

    (topLevelLoop(e), idToVarId)
  }

  def fromFile(f: String): CNF = fromFile(new File(f))

  def fromFile(f: File): CNF = {
    CNF(Source.fromFile(f).getLines.filter(line =>
      !line.trim.startsWith("c") &&
      !line.trim.startsWith("p") &&
      !line.trim.startsWith("0") &&
      !line.trim.startsWith("%") &&
      line.trim.nonEmpty
    ).map(line =>
      Clause(line.replace("\t", " ").split(" ").filter(s => s != "0" && s.trim.nonEmpty).map(s => Lit(s.toInt)).toArray)
    ).toArray, f.toString)
  }

  def random(numVars: Int, maxClauses: Int, maxClauseLength: Int): CNF = {
    CNF((0 to Random.nextInt(maxClauses)).map(i =>
      Clause((0 to Random.nextInt(maxClauseLength-2) + 1).map(j =>
        Lit((Random.nextInt(numVars) + 1) * bool2Int(Random.nextBoolean))
      ).toArray)
    ).toArray)
  }

}
