import Utils._
import Tseytin._
import scala.annotation.tailrec

object SatSolver {

  import Solver._

  object Solver {
    def apply(cnf: CNF): Solver = {
      new {
        val numVars: Int = cnf.maxId
        val clauses: Array[Clause] = cnf.clauses
        val numClauses: Int = clauses.size
      } with Solver
    }

    def checkSat(e: Expr): Option[Map[Identifier, Boolean]] = {
      val (cnf, idToVar) = CNF.makeCNF(tseytin(e))
      checkSat(cnf).map(array =>
        idToVar.mapValues(x => int2Bool(array(x.id)))
      )
    }

    def checkSat(cnf: CNF): Option[Array[Int]] = {
      val solver = Solver(cnf)
      solver.init()

      while (true) {
        // We propagate whatever can be propagated
        solver.propagate()

        if (!solver.hasConflict) {
          if (solver.allAssigned) {
            // If we don't have a conflict and all variables are assigned,
            // we have found a satisfying assignment.
            return Some(solver.assignments)
          } else {
            // If some variables remain unassigned, we need to make a new
            // decision.
            assert(solver.nextDecision.nonEmpty, "We should have a nextDecision here.")
            solver.decide(solver.nextDecision.get)
          }
        }
        else {
          // If we have a conflict, we need to backtrack
          solver.backtrack()
          // If `backtrack` did not set `nextDecision`, it means we have
          // explored all assignments and the formula is unsat.
          if (solver.nextDecision.isEmpty)
            return None
        }
      }

      sys.error("Cannot exit while(true) loop.")
    }
  }

  sealed abstract class Origin
  case object Decision extends Origin
  case object Propagation extends Origin

  trait Solver {

    /** These constants do not change during solving **/
    val numVars: Int
    val clauses: Array[Clause]
    val numClauses: Int

    require(numVars > 0)


    /** These are mutable and are modified during solving **/

    // The stack of assignments that we have made so far (can be backtracked
    // by the `backtrack` function).
    // For each literal, we remember whether it was set by a decision or by a
    // propagation.
    var trail: List[(Lit, Origin)] = List()

    // `assignments(i) =  0` means variable i is not assigned
    // `assignments(i) =  1` means variable i is assigned to `true`
    // `assignments(i) = -1` means variable i is assigned to `false`
    val assignments: Array[Int] = Array.ofDim(numVars + 1)

    // For each clause, we maintain the two literals that are watched
    var watchedLiterals: Array[Set[Lit]] = Array.ofDim(numClauses)

    // `watchingClauses(x)` returns the set of clauses indices that watch `x`.
    val watchingClauses: Array[Set[Int]] = Array.fill(numVars + 1)(Set())

    // This variable is set to `true` when a conflict is found
    var conflict: Boolean = false

    // This variable sets the next decision to make
    var nextDecision: Option[Lit] = Some(VarId(1).toLiteral(true))

    // During initialization,
    // * we make sure there are no empty clauses, and
    // * we initialize `watchedLiterals` and `watchingClauses`.
    def init(): Unit = {
      // No empty clause
      for (i <- 0 to clauses.length - 1) {
        assert(!clauses(i).isEmpty, "Empty clause means unsat")
      }

      // initialization of `watchedLiterals` and `watchingClauses`
      for (i <- 0 to clauses.length - 1) {
        watchedLiterals(i) = clauses(i).literals.toSet.take(2)
        for (lit <- watchedLiterals(i)) {
          watchingClauses(lit.getVar.id) = watchingClauses(lit.getVar.id) + i
        }
      }
    }

    def assignment(x: VarId): Int = assignments(x.id)

    def assume(lit: Lit, origin: Origin): Unit = {
      require (unassigned(lit), "We should only assign non-assigned literals")
      trail = (lit, origin) :: trail
      assignments(lit.getVar.id) = lit.positiveValue
    }

    def unassign(lit: Lit): Unit = {
      assignments(lit.getVar.id) = 0
    }

    def hasConflict: Boolean = conflict

    def unassigned(lit: Lit): Boolean = {
      assignments(lit.getVar.id) == 0
    }

    def isWellAssigned(lit: Lit): Boolean = {
      assignments(lit.getVar.id) == lit.positiveValue
    }

    def isBadlyAssigned(lit: Lit): Boolean = {
      assignments(lit.getVar.id) == -lit.positiveValue
    }

    /** Stop watching literal `lit` in clause `i`,
      * by updating `watchingClauses` and `watchedLiterals`.
      */
    def stopWatching(i: Int, lit: Lit): Unit = {
      ???
    }

    /** Start watching literal `lit` in clause `i`,
      * by updating `watchingClauses` and `watchedLiterals`.
      **/
    def startWatching(i: Int, lit: Lit): Unit = {
      ???
    }

    def propagate(): Unit = {
      propagate((0 to clauses.length - 1).flatMap(propagateClause).toSet)
    }

    /** This function explores clause `i` to look for a literal to propagate.
      * It considers three main cases:
      *
      * - A literal of `clauses(i)` is already well assigned.
      *
      * - A single literal watched by `i` is unassigned: we consider two sub cases:
      *   + There is another literal in `clauses(i)` which is unassigned
      *     and not watched yet by clause `i`. In which case, we update the
      *     watched literals accordingly (using `stopWatching` and `startWatching`).
      *
      *   + Otherwise, we return the literal to propagate.
      *
      * - If the previous cases do not apply, we do nothing.
      *
      * Hint: Do not invoke `propagate` or use `assume` to assign literals,
      *       this must be done from within `propagate` itself.
      */
    def propagateClause(i: Int): Option[Lit] = {
      ???
    }

    /** This function takes as argument a set of literals to propagate.
      *
      * After assigning the value to a literal using `assume`, it must explore
      * the clauses watching the corresponding variable with `propagateClause`
      * to see if there are more literals to propagate.
      *
      * While propagating literals, if the value of the literal has already been
      * assigned to the opposite value, the function must set the `conflict`
      * property to `true` to mark that a conflict has been found.
      */
    def propagate(toPropagate: Set[Lit]): Unit = {
      ???
    }

    def allAssigned(): Boolean = assignments.tail.forall(_ != 0)

    def nextVar(lit: Lit): Option[VarId] = {
      val id = lit.getVar.id
      if (id < numVars) Some(VarId(id + 1))
      else None
    }

    def decide(lit: Lit): Unit = {
      if (unassigned(lit)) {
        assume(lit, Decision)
      }
      nextDecision = nextVar(lit).map(_.toLiteral(true))
    }

    // Attempt to backtrack and set the next decision to make
    // We pop the `trail` stack until we find a variable which was set to `true`
    // by a decision. When we find it, we set the next decision to be this
    // variable set to `false`.
    def backtrack(): Unit = trail match {
      case Nil =>
        // no more decisions to make, we have a conflict and we are unsat
        nextDecision = None
      case (lit, decision) :: lits =>
        trail = lits
        unassign(lit)
        if (decision == Decision && lit.isPositive) {
          conflict = false
          nextDecision = Some(-lit)
        } else
          backtrack()
    }

    override def toString: String = {
      "\n\n============================\n" +
      "Assignment:\n" +
        assignments.zipWithIndex.tail.map { case (v,i) => "  " + i + " -> " + v }.mkString("\n") + "\n" +
      "Initial clauses:\n" +
        clauses.zipWithIndex.map { case (s, i) =>
          "  " + i + " -> " + s
        }.mkString("\n") + "\n" +
      "Watched Literals:\n" +
        watchedLiterals.zipWithIndex.map { case (s, i) =>
          "  " + i + " -> " + s.toList.mkString(" ")
        }.mkString("\n") +
      "\nNext decision to make: " + nextDecision +
      "\nConflict: " + conflict +
      "\nTrail: " + trail.mkString(" ") +
      "\n============================\n"
    }

  }
}
