import Expr._
import SatSolver._
import ScalaZ3Interface._
import Tseytin._

object Tests {
  val ids: Array[Identifier] = (0 to 9).map(i => StateIdentifier(s"x$i")).toArray
  val vars: Array[Expr] = ids.map(Var)
  val bTrue: Expr = BooleanLiteral(true)
  val bFrue: Expr = BooleanLiteral(false)
  val basicExprs: Seq[Expr] = Seq(
    bFrue,
    bTrue,
    vars(0),
    And(vars(0), vars(1)),
    and(vars(0), vars(1), Not(vars(2))),
    and(vars(0), vars(1), Not(vars(0))),
    Or(vars(0), vars(1)),
    Xor(vars(0), vars(1)),
    xor(vars(0), vars(1), vars(2)),
    Xor(vars(0), vars(0)),
    Xor(vars(0), Xor(vars(1), vars(1))),
    Not(Or(bTrue, vars(0))),
    Implies(vars(0), vars(1)),
    Equal(vars(0), vars(1)),
    Not(vars(0)),
    And(vars(0), Not(vars(1))),
    and(
      Not(vars(0)),
      And(vars(0), vars(0))
    ),
    and(
      Or(vars(0), vars(1)),
      Or(vars(0), Not(vars(1))),
      Or(Not(vars(0)), vars(1)),
    ),
    and(
      Or(vars(0), vars(1)),
      Or(vars(0), Not(vars(1))),
      Or(Not(vars(0)), vars(1)),
      Or(Not(vars(0)), Not(vars(1))),
    ),
    Implies(
      vars(0),
      Not(vars(0))
    ),
    Implies(
      vars(0),
      Xor(vars(0), vars(0))
    ),
    Implies(
      vars(0),
      Xor(vars(1), vars(1))
    )
  )

  // This function checks that the result of Tseytin's transformation on
  // expression `e` is indeed in CNF, and that `z3CheckSat` says "satisfiable"
  // for the result if and only if it says "satisfiable" for the original
  // expression.
  def testTseytin(e: Expr, print: Boolean): Boolean = {
    if (print) {
      println("Checking Tseytin's transformation on: " + e)
    }
    val r1 = z3CheckSat(e)
    val te = tseytin(e)
    assert(te.isCNF, "The result of Tseytin's transformation should be in CNF form.")
    val r2 = z3CheckSat(te)
    if (r1.isEmpty != r2.isEmpty) {
      println("Found an error while running `z3CheckSat` on Tseytin's transformation")
      println("On expression:\n" + e)
      println("The transformation is:\n" + te)
      println("\n`z3CheckSat` on the original expression returned:")
      println(r1)
      println("\nWhile `z3CheckSat` on Tseytin's transformation returned:")
      println(r2)
      false
    } else {
      true
    }
  }

  def nameOrContent(cnf: CNF): String = {
    if (cnf.name.nonEmpty) cnf.name
    else cnf.toString
  }

  // This function checks that `Solver.checkSat` and `z3CheckSat` return the
  // same result.
  def testCheckSatCNF_Z3(cnf: CNF): Boolean = {
    if (cnf.name.nonEmpty)
      println("testCheckSatCNF_Z3: " + cnf.name)
    val r1 = Solver.checkSat(cnf)
    val r2 = z3CheckSat(cnf.toExpr)
    if (r1.isEmpty != r2.isEmpty) {
      println("Found an error on `SatSolver.checkSat`")
      println("On cnf:\n" + nameOrContent(cnf))
      println("\n`SatSolver.checkSat` returned:")
      println(r1.map(_.mkString("\n")))
      println("\nWhile `z3CheckSat` on cnf.toExpr returned:")
      println(r2)
      false
    } else {
      true
    }
  }

  // This function checks that `Solver.checkSat` says "satisfiable"
  def testCheckSatCNF_Sat(cnf: CNF): Boolean = {
    if (cnf.name.nonEmpty)
      println("testCheckSatCNF_Sat: " + cnf.name)
    val r = Solver.checkSat(cnf)
    if (r.isEmpty) {
      println("Found an error on `SatSolver.checkSat`")
      println("On cnf:\n" + nameOrContent(cnf))
      println("\n`SatSolver.checkSat` returned unsat.")
      println("\nWhile the formula is satisfiable.")
      false
    } else {
      true
    }
  }

  // This function checks that `Solver.checkSat` says "unsatisfiable"
  def testCheckSatCNF_Unsat(cnf: CNF): Boolean = {
    if (cnf.name.nonEmpty)
      println("testCheckSatCNF_Unsat: " + cnf.name)
    val r = Solver.checkSat(cnf)
    if (r.nonEmpty) {
      println("Found an error on `SatSolver.checkSat`")
      println("On cnf:\n" + nameOrContent(cnf))
      println("\n`SatSolver.checkSat` returned a satisfying assignment.")
      println(r.map(_.mkString("\n")))
      println("\nWhile the formula is unsatisfiable.")
      false
    } else {
      true
    }
  }

  def testThese[T](toTest: Iterable[T], name: String, f: T => Boolean): Boolean = {
    toTest.foldLeft(true)((acc, e) => acc && {
      val r = f(e)
      if (r)
        println("  ==> OK")
      else
        println("  ==> Error")
      r
    })
  }

  def testRandom[T](n: Int, random: => T, name: String, f: T => Boolean): Boolean = {
    for (i <- 1 to n) {
      if (!f(random))
        return false
      if (i % 100 == 0)
        println(s"Did $i random (successful) tests for `$name`")
    }
    return true
  }

}
